
#ifndef ___GL_OBJECT_H___
#define ___GL_OBJECT_H___

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
// データ変換ワーニングの無効化
#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
#endif

//#ifdef   FX_GL_BETA_2
//#include "gl.h"
//#else
#include <GL/gl.h>
//#endif

#include <GL/glut.h>
#include <iostream.h>

#include "Math3D.H"
#include "MString.H"
#include "PNMIO.H"

// SGIS_multitexture
#if defined WIN32 && !defined TEXTURE0_SGIS 
#define TEXTURE0_SGIS	0x835E
#define TEXTURE1_SGIS	0x835F
#endif

#ifdef WIN32
class GLExtensionsProcPointer
{
	// extensions string list
	String extensions ;

public:
	// extension flags
	int SGIS_multitexture ;

	// SGIS_multitexture procedure pointers
	PROC glSelectTextureCoordSetSGIS ;
	PROC glSelectTextureSGIS ;
	PROC glMTexCoordPointerSGIS ;
	PROC glMTexCoord1fSGIS ;
	PROC glMTexCoord1fvSGIS ;
	PROC glMTexCoord2fSGIS ;
	PROC glMTexCoord2fvSGIS ;
	PROC glMTexCoord3fSGIS ;
	PROC glMTexCoord3fvSGIS ;

	// constructor
	GLExtensionsProcPointer() ;
} ;
#endif

class Vector2d ;

// イメージフォーマットを RGB から RGBA に
int RGBtoRGBA(TRUEIMAGE *image, GLfloat alpha = 1.0f) ;

// イメージガンマ補正
void GammaCollect(TRUEIMAGE *image, float gamma, unsigned int ix = 0, unsigned int iy = 0, unsigned int width = 0, unsigned int height = 0) ;
// イメージの明るさ変更
void ChangeBright(TRUEIMAGE *image, float gamma, unsigned int ix = 0, unsigned int iy = 0, unsigned int width = 0, unsigned int height = 0) ;

// イメージの上下左右反転
void ReverseCoordinates(TRUEIMAGE *image, int mirrorFlag, int flipFlag) ;

// 視野ピラミッド作成（ジッタ情報含）
void     Perspective(GLfloat scr_w, GLfloat scr_h, GLfloat n_clip, GLfloat f_clip,
				 int win_w = NONE, int win_h = NONE, GLfloat px = 0.0f, GLfloat py = 0.0f) ;

Vector2d *CreateSampleJitter(int sample) ;
void     DeleteSampleJitter(Vector2d *jitter) ;

void     SetMaterial(GLenum surface, GLfloat r, GLfloat g, GLfloat b, GLfloat a = 1.0f, GLfloat metal = 0.0f,
					 GLfloat specular = 1.5f, GLfloat shininess = 64.0f, GLint mirror = FALSE, GLfloat m_ratio = 0.0f, GLfloat amb = NONE, GLenum material = GLenum(NONE)) ;
void     SetLight(GLenum light, GLfloat pos[4], GLfloat color[4], GLfloat amb = 0.1f, GLint unitFlag = TRUE, GLint specularFlag = TRUE) ;

// 文字列描画
void DrawString(String& str, GLfloat x = 0.0f, GLfloat y = 0.0f, GLfloat z = 0.0f, void *font = GLUT_BITMAP_9_BY_15) ;

// 文字列描画
void DrawString(String& str, void *font = GLUT_BITMAP_9_BY_15) ;

// １文字描画
inline void DrawCharacter(char chr, GLfloat x = 0.0f, GLfloat y = 0.0f, GLfloat z= 0.0f, void *font = GLUT_BITMAP_9_BY_15)
{
	glRasterPos3d(x, y, z) ;
	glutBitmapCharacter(font, chr) ;
}

class Vector2d
{
public:
	GLfloat x, y ;

	Vector2d() {} ;
	Vector2d(GLfloat vx, GLfloat vy) { x = vx ; y = vy ; }
	~Vector2d() {} ;

	Vector2d& operator ()(GLfloat vx, GLfloat vy) { x = vx ; y = vy ; return *this ; }
} ;


class Vector3d
{
public:
    GLfloat v[3] ;

    Vector3d() {} ;
    Vector3d(GLfloat x, GLfloat y, GLfloat z) { v[0] = x ; v[1] = y ; v[2] = z ; }
    Vector3d(const Vector3d& ov) { v[0] = ov.v[0] ; v[1] = ov.v[1] ; v[2] = ov.v[2] ; }
    Vector3d(GLfloat *ov) { v[0] = ov[0] ; v[1] = ov[1] ; v[2] = ov[2] ; }

    // アクセサ
    GLfloat& X() { return v[0] ; }
    GLfloat& Y() { return v[1] ; }
    GLfloat& Z() { return v[2] ; }

    const GLfloat& X() const { return v[0] ; }
    const GLfloat& Y() const { return v[1] ; }
    const GLfloat& Z() const { return v[2] ; }

    // メソッド
    Vector3d& Unit() ;
//    friend const Vector3d Unit(Vector3d& vec) ;
    friend Vector3d Unit(const Vector3d& vec) ;
//    friend const Vector3d UnitNormal(Vector3d& v1, Vector3d& v2) ;
//    friend const Vector3d UnitNormal(Vector3d& v1, const Vector3d& v2) ;
//    friend const Vector3d UnitNormal(const Vector3d& v1, Vector3d& v2) ;
    friend Vector3d UnitNormal(const Vector3d& v1, const Vector3d& v2) ;
    Vector3d& Transform(const GLfloat m[16]) ;
    Vector3d& Transform3x3(const GLfloat m[16]) ;
    float Abs() const { return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]) ; }

    // オペレータ
    const Vector3d& operator ()(GLfloat x, GLfloat y, GLfloat z)
      { v[0] = x ; v[1] = y ; v[2] = z ; return *this ; }

//    friend const Vector3d operator |(Vector3d& v1, Vector3d& v2) ;
//    friend const Vector3d operator |(Vector3d& v1, const Vector3d& v2) ;
//    friend const Vector3d operator |(const Vector3d& v1, Vector3d& v2) ;
    friend Vector3d operator |(const Vector3d& v1, const Vector3d& v2) ;

    // メンバ関数 X(),Y(),Z() が const でないため、引数は const に出来ない
//    friend int operator ==(Vector3d& v1, Vector3d& v2)
//      { return (v1.v[0] == v2.v[0] && v1.v[1] == v2.v[1] && v1.v[2] == v2.v[2]) ; }
//    friend int operator ==(Vector3d& v1, const Vector3d& v2)
//      { return (v1.v[0] == v2.v[0] && v1.v[1] == v2.v[1] && v1.v[2] == v2.v[2]) ; }
//    friend int operator ==(const Vector3d& v1, Vector3d& v2)
//      { return (v1.v[0] == v2.v[0] && v1.v[1] == v2.v[1] && v1.v[2] == v2.v[2]) ; }
    friend int operator ==(const Vector3d& v1, const Vector3d& v2)
      { return (v1.v[0] == v2.v[0] && v1.v[1] == v2.v[1] && v1.v[2] == v2.v[2]) ; }

//    friend int operator !=(Vector3d& v1, Vector3d& v2)
//      { return !(v1 == v2) ; }
//    friend int operator !=(Vector3d& v1, const Vector3d& v2)
//      { return !(v1 == v2) ; }
//    friend int operator !=(const Vector3d& v1, Vector3d& v2)
//      { return !(v1 == v2) ; }
    friend int operator !=(const Vector3d& v1, const Vector3d& v2)
      { return !(v1 == v2) ; }

//    Vector3d& operator =(Vector3d& ov)
//      { v[0] = ov.v[0] ; v[1] = ov.v[1] ; v[2] = ov.v[2] ; return *this ; }

//    const Vector3d& operator =(Vector3d& ov)
//      { v[0] = ov.v[0] ; v[1] = ov.v[1] ; v[2] = ov.v[2] ; return *this ; }
    const Vector3d& operator =(const Vector3d& ov)
      { v[0] = ov.v[0] ; v[1] = ov.v[1] ; v[2] = ov.v[2] ; return *this ; }

//    const Vector3d& operator =(GLfloat ov[3])
//      { v[0] = ov[0] ; v[1] = ov[1] ; v[2] = ov[2] ; return *this ; }
    const Vector3d& operator =(const GLfloat ov[3])
      { v[0] = ov[0] ; v[1] = ov[1] ; v[2] = ov[2] ; return *this ; }

//    const Vector3d& operator +=(Vector3d& ov)
//      { v[0] += ov.v[0] ; v[1] += ov.v[1] ; v[2] += ov.v[2] ; return *this ; }
    const Vector3d& operator +=(const Vector3d& ov)
      { v[0] += ov.v[0] ; v[1] += ov.v[1] ; v[2] += ov.v[2] ; return *this ; }

//    friend const Vector3d operator +(Vector3d& v1, Vector3d& v2)
//      { return Vector3d(v1.v[0]+v2.v[0], v1.v[1]+v2.v[1], v1.v[2]+v2.v[2]) ; }
//    friend const Vector3d operator +(Vector3d& v1, const Vector3d& v2)
//      { return Vector3d(v1.v[0]+v2.v[0], v1.v[1]+v2.v[1], v1.v[2]+v2.v[2]) ; }
//    friend const Vector3d operator +(const Vector3d& v1, Vector3d& v2)
//      { return Vector3d(v1.v[0]+v2.v[0], v1.v[1]+v2.v[1], v1.v[2]+v2.v[2]) ; }
    friend Vector3d operator +(const Vector3d& v1, const Vector3d& v2)
      { return Vector3d(v1.v[0]+v2.v[0], v1.v[1]+v2.v[1], v1.v[2]+v2.v[2]) ; }

//    const Vector3d& operator -=(Vector3d& ov)
//      { v[0] -= ov.v[0] ; v[1] -= ov.v[1] ; v[2] -= ov.v[2] ; return *this ; }
    const Vector3d& operator -=(const Vector3d& ov)
      { v[0] -= ov.v[0] ; v[1] -= ov.v[1] ; v[2] -= ov.v[2] ; return *this ; }

//    friend const Vector3d operator -(Vector3d& v1, Vector3d& v2)
//      { return Vector3d(v1.v[0]-v2.v[0], v1.v[1]-v2.v[1], v1.v[2]-v2.v[2]) ; }
//    friend const Vector3d operator -(Vector3d& v1, const Vector3d& v2)
//      { return Vector3d(v1.v[0]-v2.v[0], v1.v[1]-v2.v[1], v1.v[2]-v2.v[2]) ; }
//    friend const Vector3d operator -(const Vector3d& v1, Vector3d& v2)
//      { return Vector3d(v1.v[0]-v2.v[0], v1.v[1]-v2.v[1], v1.v[2]-v2.v[2]) ; }
    friend Vector3d operator -(const Vector3d& v1, const Vector3d& v2)
      { return Vector3d(v1.v[0]-v2.v[0], v1.v[1]-v2.v[1], v1.v[2]-v2.v[2]) ; }

    const Vector3d& operator *=(float d)
      { v[0] *= d ; v[1] *= d ; v[2] *= d ; return *this ; }
//    friend const Vector3d operator *(Vector3d& v, float d)
//      { return Vector3d(v.v[0] * d, v.v[1] * d, v.v[2] * d) ; }
    friend Vector3d operator *(const Vector3d& v, float d)
      { return Vector3d(v.v[0] * d, v.v[1] * d, v.v[2] * d) ; }

    const Vector3d& operator /=(float d)
      { v[0] /= d ; v[1] /= d ; v[2] /= d ; return *this ; }
//    friend const Vector3d operator /(Vector3d& v, float d)
//      { return Vector3d(v.v[0] / d, v.v[1] / d, v.v[2] / d) ; }
    friend Vector3d operator /(const Vector3d& v, float d)
      { return Vector3d(v.v[0] / d, v.v[1] / d, v.v[2] / d) ; }

    static const Vector3d white ;
} ;

ostream& operator <<(ostream& os, Vector3d& vec) ;

inline GLfloat* VectorToArray(GLfloat v[3], const Vector3d& vec)
{
	v[0] = vec.X() ;
	v[1] = vec.Y() ;
	v[2] = vec.Z() ;

	return v ;
}


class VPolygon
{
public:
    Vector3d vertex[4], normal[4], pNormal ;

	VPolygon& operator=(const VPolygon& v)
	{
		for (int i = 0 ; i < 4 ; i ++)
		{
			vertex[i] = v.vertex[i] ;
			normal[i] = v.normal[i] ;
		}
		pNormal = v.pNormal ;

		return *this ;
	}
} ;


class Object
{
public:
    int     nPolygons, fType, eType ;
    VPolygon *polygon ;

    Object(int nPol = 0)
	{
		if (nPol >= 0) polygon = new VPolygon[nPol] ;
		else { polygon = NULL ; nPol = 0 ; }
		nPolygons = nPol ;
	}

    ~Object() { delete polygon ; }

	Object& NPolygons(int nPol)
	{
		if (polygon) delete polygon ;
		if (nPol < 0) nPol = 0 ;
		polygon = new VPolygon[nPol] ;
		nPolygons = nPol ;
		return *this ;
	}

    Object& MakePolygon(int polNo, int v1, int v2, int v3, const GLfloat data[][3], const GLfloat m[16] = NULL) ;
    Object& MakePolygon(int polNo, int v1, int v2, int v3, int v4, const GLfloat data[][3], const GLfloat m[16] = NULL) ;
    Object& AverageNormal(int pNo, int vNo, int pNo1, int pNo2, int pNo3 = -1, int pNo4 = -1, int pNo5 = -1, int pNo6 = -1) ;
	Object& DrawTriangle(int p = 0, int depth = FALSE, int average = FALSE, int flatFlag = FALSE) ;
	Object& DrawTriangles(int depth = FALSE, int average = TRUE, int flatFlag = FALSE) ;
} ;


class RollingObject: public Object
{
public: 
	int fType, eType, uDiv, vDiv ;

	// コンストラクタ
	RollingObject(int nPol = 0): Object(nPol) { uDiv = vDiv = 0 ; }

	RollingObject& NPolygons(int nPol) { Object::NPolygons(nPol) ; return *this ; }
	RollingObject& MakeRollingPolygons(int u, int v, GLfloat data[][2], GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f) ;
	static int RollingObject::VFlag(int v, int ud, int vFlag, int vFlags[][2]) ;
	RollingObject& AverageNormals(int uFlag = TRUE, int vFlag = TRUE, int vFlags[][2] = NULL) ; //, int vNo = 0) ;

	// それぞれ、三角ポリゴン（頂点がオブジェクトの中心１点）の場合のみ対応
	RollingObject& TopNormal(GLfloat x = 0.0f, GLfloat y = 0.0f, GLfloat z = 1.0f) ;
	RollingObject& BottomNormal(GLfloat x = 0.0f, GLfloat y = 0.0f, GLfloat z = -1.0f) ;

	void DrawObject() ;
} ;


class PositivePolyhedron: public Object
{
public:
    int depth, average ;

    PositivePolyhedron(int nPols): Object(nPols) {} ;

	PositivePolyhedron& Smooth() ;
	PositivePolyhedron& Depth(int d) { depth = d ; return *this ; }
	PositivePolyhedron& Average(int b) { average = b ; return *this ; }

	PositivePolyhedron& Draw(int d = NONE, int a = NONE, int flatFlag = FALSE)
	{
		if (d >= 0) depth   = d ;
		if (a >= 0) average = a ;
		DrawTriangles(depth, average, flatFlag) ;
		return *this ;
	}
} ;


class Octahedron: public PositivePolyhedron
{
public:
	Octahedron& Make(GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, int d = NONE, int a = NONE, int smooth = NONE) ;
    Octahedron(GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, int d = FALSE, int a = TRUE, int smooth = FALSE): PositivePolyhedron(8)
	  { Make(x,y,z, d, a, smooth) ; }

	static void DrawPolyhedron(GLfloat r = 1.0) ;

	static GLfloat o_v[][3] ;
	static int     o_o[][3] ;
} ;


class Icosahedron: public PositivePolyhedron
{
public:
	Icosahedron& Make(GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, int d = NONE, int a = NONE, int smooth = NONE) ;
    Icosahedron(GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, int d = FALSE, int a = TRUE, int smooth = FALSE): PositivePolyhedron(20)
	  { Make(x,y,z, d, a, smooth) ; }

	static void DrawPolyhedron(GLfloat r = 1.0) ;

	static GLfloat X, Z, i_v[][3] ;
	static int  i_o[][3] ;
} ;


// 屈折ベクトル、鏡面反射係数計算クラス
class Refraction
{
public:
	GLfloat  specular ;
	Vector3d refraction ;

	Refraction& operator =(const Refraction& r)
	{
		specular = r.specular ;
		refraction = r.refraction ;

		return *this ;
	}

	// n1, n2, eye, normal から specular および refraction を計算
	// 反射係数や透過係数の計算も行う
	Refraction& CalculateRefraction(GLfloat n, Vector3d&eye, Vector3d& normal)
        {
		float n2 = n ; // n2 / n1 ;
		float cos1 = -(eye.X() * normal.X() +
			       eye.Y() * normal.Y() +
			       eye.Z() * normal.Z()) ;
		float sin1 = 1.0f - cos1 * cos1 ;
		if (sin1 < 0.0f) specular = 1.0f ; // 完全鏡面反射
		else
		{
			float cos2 = sqrt(1.0f - sin1 / (n2 * n2)) ;
			float n2cos2cos1 = n2 * cos2 - cos1 ;

			float kss = Pow(n2cos2cos1 / (n2 * cos2 + cos1)) ;
			float ksp = Pow((n2 * cos1 - cos2) / (n2 * cos1 + cos2)) ;
			specular = 0.5f * (kss + ksp) ;

			refraction((eye.X() - normal.X() * n2cos2cos1) / n2,
				   (eye.Y() - normal.Y() * n2cos2cos1) / n2,
				   (eye.Z() - normal.Z() * n2cos2cos1) / n2) ;
		}

		return *this ;
	}

	// n1, n2, eye, normal から refraction を計算
	// 反射係数や透過係数の計算なしバージョン
	Refraction& CalculateRefraction2(GLfloat n, Vector3d&eye, Vector3d& normal)
        {
		float n2 = n ; // n2 / n1 ;
		float cos1 = -(eye.X() * normal.X() +
			       eye.Y() * normal.Y() +
			       eye.Z() * normal.Z()) ;
		float sin1_sq = 1.0f - cos1 * cos1 ;
		float root = n2 * n2 - sin1_sq ;
		if (root < 0.0f) root = 0.0f ;
		float sqrtn2sin = sqrt(root) - cos1 ;

		refraction = (eye - normal * sqrtn2sin) / n2 ;

//		cout << refraction << "   length: " << refraction.Abs() << endl ;

		return *this ;
	}


	// eye, normal から specular を計算
	// 反射係数や透過係数の計算を行うだけで、
	// 実際には視線ベクトルは屈折させない
	// フェイクを行う場合、n は入射角による反射透過の変化率を指定
	Refraction& CalculateFakeRefraction(GLfloat n, Vector3d&eye, Vector3d& normal)
        {
		float cos1 = -(eye.X() * normal.X() +
			       eye.Y() * normal.Y() +
			       eye.Z() * normal.Z()) ;

		// 視線ベクトルと法線ベクトルの角度から鏡面反射率を計算（もろフェイク）
		if (n == 1.0f) specular = 1.0f - cos1 ;
		else specular = 1.0f - Pow(cos1, n) ;
		return *this ;
	}

	Refraction() { } ;
} ;


// 屈折マッピング使用時の各頂点情報
// 頂点座標、法線、テクスチャ座標、鏡面反射係数、透過係数
class RefractVertex
{
public:
	Vector3d   vertex, normal, texture, specular_color, refraction_color ;
	GLfloat    n ;
	Refraction refraction ;

	RefractVertex() {} ;
	RefractVertex(Vector3d& v, Vector3d& norm, GLfloat n2)
        {
		vertex = v ;
		normal = norm ;
		n = n2 ;
	}

	RefractVertex& operator =(const RefractVertex& v)
	{
		vertex = v.vertex ;
		normal = v.normal ;
		texture = v.texture ;
		specular_color = v.specular_color ;
		refraction_color = v.refraction_color ;
		n = v.n ;
		refraction = v.refraction ;

		return *this ;
	}

	RefractVertex(const RefractVertex& v)
	{
		*this = v ;
	}

	// 透過率、反射率を考慮に入れた計算
	RefractVertex& CalculateVertex(const GLfloat matrix[16], const Vector3d& r_c, const Vector3d& s_c = Vector3d::white)
	{
		Vector3d eye, norm ;
		eye = vertex ;
		eye.Transform(matrix) ;

		if (eye.Z() >= 0.0f) return *this ;
		eye.Unit() ;

		norm = normal ;
		norm.Transform3x3(matrix) ;

		// 屈折ベクトル、反射係数計算
		refraction.CalculateRefraction(n, eye, norm) ;

		specular_color = s_c * refraction.specular ;
		refraction_color = r_c * (1.0f - refraction.specular) ;

		// 屈折ベクトルからテクスチャ座標計算
		GLfloat m = 2.0f * sqrt(Pow(refraction.refraction.X()) +
					Pow(refraction.refraction.Y()) +
					Pow(refraction.refraction.Z() + 1.0f)) ;

		if (m >= 0.0f)
			texture(refraction.refraction.X() / m + 0.5f,
				refraction.refraction.X() / m + 0.5f, 0.0f) ;
		else texture(0.0f, 0.5f, 0.0f) ;

		return *this ;
	}

	// 透過率、反射率は無関係（描画時に色は指定しない）
	// アプリケーション側で glColor*() で指定
	RefractVertex& CalculateVertex(const GLfloat matrix[16])
	{
		Vector3d eye, norm ;
		eye = vertex ;
		eye.Transform(matrix) ;

		if (eye.Z() >= 0.0f) return *this ;
		eye.Unit() ;

		norm = normal ;
		norm.Transform3x3(matrix) ;

		// 屈折ベクトル計算
		refraction.CalculateRefraction2(n, eye, norm) ;

		// 屈折ベクトルからテクスチャ座標計算
		GLfloat m = 2.0f * sqrt(Pow(refraction.refraction.X()) +
					Pow(refraction.refraction.Y()) +
					Pow(refraction.refraction.Z() + 1.0f)) ;
		if (m > 0.0f)
			texture(refraction.refraction.X() / m + 0.5f,
				refraction.refraction.X() / m + 0.5f, 0.0f) ;
		else texture(0.0f, 0.5f, 0.0f) ;

		return *this ;
	}


	// フェイク屈折...
	// 透過率、反射率を考慮に入れた計算
	RefractVertex& CalculateFakeVertex(const GLfloat matrix[16], const Vector3d& r_c, const Vector3d& s_c = Vector3d::white)
	{
		Vector3d eye, norm ;
		eye = vertex ;
		eye.Transform(matrix) ;

		if (eye.Z() >= 0.0f) return *this ;
		eye.Unit() ;

		norm = normal ;
		norm.Transform3x3(matrix) ;

		// 反射係数計算
		refraction.CalculateFakeRefraction(n, eye, norm) ;

		specular_color = s_c * refraction.specular ;
		refraction_color = r_c * (1.0f - refraction.specular) ;

		// 屈折ベクトル（この場合は視線ベクトルそのまま）からテクスチャ座標計算
		GLfloat m = 2.0f * sqrt(Pow(eye.X()) +
					Pow(eye.Y()) +
					Pow(eye.Z() + 1.0f)) ;

/*
		if (m >= 0.0f)
			texture(eye.X() / m + 0.5f,
				eye.X() / m + 0.5f, 0.0f) ;
		else texture(0.0f, 0.5f, 0.0f) ;
*/
		texture(eye.X() / m + 0.5f,
			eye.X() / m + 0.5f, 0.0f) ;

		return *this ;
	}

	//
	RefractVertex& CalculateFakeVertex(const GLfloat matrix[16])
	{
		Vector3d eye, norm ;
		eye = vertex ;
		eye.Transform(matrix) ;

		if (eye.Z() >= 0.0f) return *this ;
		eye.Unit() ;

		// 屈折ベクトル（この場合は視線ベクトルそのまま）からテクスチャ座標計算
		GLfloat m = 2.0f * sqrt(Pow(eye.X()) +
					Pow(eye.Y()) +
					Pow(eye.Z() + 1.0f)) ;

		texture(eye.X() / m + 0.5f,
			eye.X() / m + 0.5f, 0.0f) ;

		return *this ;
	}
} ;


// 屈折あり回転体オブジェクト
class RefractRollingObject : public RollingObject
{
public:
	RefractVertex *verticies ;

	// コンストラクタ
	RefractRollingObject(int nPol = 0): RollingObject(nPol) { verticies = NULL ; }

	// デストラクタ 頂点単位データが生成されているなら解放
	~RefractRollingObject() { if (verticies) delete verticies ; }

	// 頂点データ配列アクセサ（エラーチェック等一切なし）
	RefractVertex& Vertex(int u, int v) { return verticies[v * (uDiv + 1) + u] ; }

	// 屈折とか用の頂点単位のデータ作成
	// すでに回転体のポリゴンデータが存在していなければならない
	// 現段階では MakeRollingObject() 内からのみコールされる
	RefractRollingObject& MakeRefractVerticies(GLfloat n = 1.5f)
	{
		if (verticies) delete verticies ;
		verticies = new RefractVertex[(uDiv + 1) * vDiv] ;

		int u, v ;
		for (v = 0 ; v < vDiv - 1 ; v ++)
		{
			int pNo = v ;

			for (u = 0 ; u <= uDiv ; u ++)
			{
				if (u == uDiv) pNo = v ;

				// 念のため初期化
				Vertex(u, v).texture(0.0f, 0.0f, 0.0f) ;
				Vertex(u, v).specular_color(1.0f, 1.0f, 1.0f) ;
				Vertex(u, v).refraction_color(0.0f, 0.0f, 0.0f) ;

				// 屈折率
				Vertex(u, v).n = n ;

				// 座標および法線ベクトル
				Vertex(u, v).vertex = polygon[pNo].vertex[0] ;
				Vertex(u, v).normal = polygon[pNo].normal[0] ;

				pNo += vDiv - 1 ;
			}
		}

		// 底の頂点データ
		// 念のため初期化
		Vertex(0, vDiv - 1).texture(0.0f, 0.0f, 0.0f) ;
		Vertex(0, vDiv - 1).specular_color(1.0f, 1.0f, 1.0f) ;
		Vertex(0, vDiv - 1).refraction_color(0.0f, 0.0f, 0.0f) ;

		// 屈折率
		Vertex(0, vDiv - 1).n = n ;

		// 座標および法線ベクトル
		Vertex(0, vDiv - 1).vertex = polygon[vDiv - 1].vertex[0] ;
		Vertex(0, vDiv - 1).normal = polygon[vDiv - 1].normal[0] ;

		return *this ;
	}

	// 回転体ポリゴンと、屈折とか用の頂点データ作成
	RefractRollingObject& MakeRollingPolygons(int u, int v, GLfloat data[][2], GLfloat n = 1.5f, GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f)
	{
		// 回転体ポリゴン生成
		RollingObject::MakeRollingPolygons(u,v, data, x,y,z) ;

		// とりあえず面倒だから、屈折オブジェクトは全てスムーズで
		// 上下のポリゴンは三角形（トップとボトム頂点がz軸中央に１つずつだけ）
		AverageNormals() ;
		TopNormal() ;
		BottomNormal() ;

		// 屈折関係用頂点単位のデータ作成
		MakeRefractVerticies(n) ;

		return *this ;
	}

	// 屈折関係データを計算しながら OpenGL描画
	// 透過率、指定した色を考慮して描画
	RefractRollingObject& DrawRefractObject(const Vector3d& r_c, const Vector3d& s_c = Vector3d::white) ;

	// 屈折関係データを計算しながら OpenGL描画
	// 透過率、色は無視（アプリケーション側が自前で glColor*() を実行）
	RefractRollingObject& DrawRefractObject() ;

	// 我流フェイク屈折マップしながら OpenGL描画（屈折率無関係）
	// 透過率、指定した色を考慮して描画
	RefractRollingObject& DrawFakeRefractObject(const Vector3d& r_c, const Vector3d& s_c = Vector3d::white) ;

	// 我流フェイク屈折マップしながら OpenGL描画（屈折率無関係）
	// 透過率、色も無視（アプリケーション側が自前で glColor*() を実行）
	RefractRollingObject& DrawFakeRefractObject() ;
} ;


class WineGlass: public RollingObject
{
public:
	enum { N_VERTICIES = 40 } ;
	static GLfloat vertexes[N_VERTICIES][2] ;

	WineGlass(int divide = 32, GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, GLfloat r = 1.0f, GLfloat g = 1.0f, GLfloat b = 1.0f)
	: RollingObject(divide * (N_VERTICIES - 1))
	{
		MakeRollingPolygons(divide, N_VERTICIES, vertexes, x / 19.5f, y / 19.5f, z / 19.5f) ;
		AverageNormals(TRUE, TRUE) ;
		TopNormal() ;
		BottomNormal() ;
	}

	static void DrawPolyhedron(int divide = 32, GLfloat r = 1.0) ;
	void DrawObject() ;
} ;

class RefractWineGlass: public RefractRollingObject
{
public:

	RefractWineGlass(int divide = 32, GLfloat n = 1.5f, GLfloat x = 1.0f, GLfloat y = 1.0f, GLfloat z = 1.0f, GLfloat r = 1.0f, GLfloat g = 1.0f, GLfloat b = 1.0f)
	: RefractRollingObject(divide * (WineGlass::N_VERTICIES - 1))
	{
		MakeRollingPolygons(divide, WineGlass::N_VERTICIES, WineGlass::vertexes, n, x / 19.5f, y / 19.5f, z / 19.5f) ;
//		AverageNormals(TRUE, TRUE) ;
//		TopNormal() ;
//		BottomNormal() ;
	}
} ;


// 回転ベクトル操作（単純な回転角度の蓄積: 0.0 から 360.0 をループ）
class RotateVector
{
public:
	Vector3d status, vector ;

	RotateVector(GLfloat sx = 0.0f, GLfloat sy = 0.0f, GLfloat sz = 0.0f,
		     GLfloat vx = 0.0f, GLfloat vy = 0.0f, GLfloat vz = 0.0f)
	: status(sx, sy, sz), vector(vx, vy, vz) { } ;

	void Transform()
	{
		status += vector ;

		if      (status.X() >= 360.0f) status.X() -= 360.0f ;
		else if (status.X() <=   0.0f) status.X() += 360.0f ;
		if      (status.Y() >= 360.0f) status.Y() -= 360.0f ;
		else if (status.Y() <=   0.0f) status.Y() += 360.0f ;
		if      (status.Z() >= 360.0f) status.Z() -= 360.0f ;
		else if (status.Z() <=   0.0f) status.Z() += 360.0f ;
	}
} ;


// テクスチャコーディネーション移動（0.0 から 1.0 をループ）
class TranslateTexture
{
public:
	Vector3d status, vector ;

	TranslateTexture(GLfloat sx = 0.0f, GLfloat sy = 0.0f, GLfloat sz = 0.0f,
			 GLfloat vx = 0.0f, GLfloat vy = 0.0f, GLfloat vz = 0.0f)
	: status(sx, sy, sz), vector(vx, vy, vz) { } ;

	void Transform()
	{
		status += vector ;

		if      (status.X() >= 1.0f) status.X() -= 1.0f ;
		else if (status.X() <= 0.0f) status.X() += 1.0f ;
		if      (status.Y() >= 1.0f) status.Y() -= 1.0f ;
		else if (status.Y() <= 0.0f) status.Y() += 1.0f ;
		if      (status.Z() >= 1.0f) status.Z() -= 1.0f ;
		else if (status.Z() <= 0.0f) status.Z() += 1.0f ;
	}
} ;


#endif
